
! greetings.f90 -- greetings program
!
! Each process other than rank #0 sends a message to rank #0.
!    Rank 0 outputs the messages received.
!
! Input: none.
! Output: contents of messages received by rank 0.
!
! See Parallel Programming with MPI, Pacheco 1996, chap. 3, pp. 41 & ff.
!
! Slightly modified by Henry Neeman.

      PROGRAM greetings

          IMPLICIT NONE
          INCLUDE "mpif.h"

          INTEGER,PARAMETER ::  maximum_message_length = 100
          INTEGER,PARAMETER ::  first_rank  = 0 ! Ranks 0..number_of_processes-1
          INTEGER,PARAMETER ::  master_rank = 0 ! Rank of master process
          INTEGER,PARAMETER ::  tag         = 0 ! Tag for messages

          CHARACTER (LEN = maximum_message_length) :: message ! Message string

          INTEGER :: recv_status(MPI_STATUS_SIZE) ! Return status for receive

          INTEGER :: number_of_processes  ! Number of processes in this run
          INTEGER :: my_rank              ! Rank of this process
          INTEGER :: source               ! Rank of sender
          INTEGER :: destination          ! Rank of receiver
          INTEGER :: message_send_length  ! Length of message to send
          INTEGER :: mpi_error_code       ! Error code returned by MPI call
          INTEGER :: idx

!........ Start up MPI.

          CALL MPI_Init(mpi_error_code)

!........ Query this MPI process's rank.

          CALL MPI_Comm_rank(MPI_COMM_WORLD, my_rank, mpi_error_code)

!........ Query the number of MPI processes in this run.

          CALL MPI_Comm_size(MPI_COMM_WORLD, number_of_processes,   &
      &                      mpi_error_code)

!........ Set the message string to all blank spaces.

          DO idx = 1, maximum_message_length

              message(idx:idx) = " "

          END DO !! idx

!........ What to do depends on which rank this is.
!........   Non-master rank: Create a message and send it to the
!........     master rank.
!........   Master rank: Loop over all the ranks (other than itself),
!........     receive a message from each such rank, and print
!........     each message that it receives.
!........
!........ Note: Because the calls to MPI_Recv occur in rank order,
!........   the outputs will be in rank order -- which doesn't mean
!........   that the messages from the various ranks actually
!........   arrived at the master rank in that order.

          IF (my_rank /= master_rank) THEN

!............ Create the message string for this non-master rank
!............ to send.

              WRITE (message, "(A,I4,A)")   &
     &            "Greetings from process ", my_rank, "!"

!............ Set the destination for this rank to send to.

              destination = master_rank

!............ Set the length of the message to send equal to the
!............ length of the message string, except for trailing
!............ blanks.

              message_send_length = LEN_TRIM(message)

!............ Send the message from this rank to the destination
!............ rank, using the appropriate length, of type CHARACTER,
!............ using the appropriate tag, via the default communicator.

              CALL MPI_Send(message, message_send_length,       &
     &                      MPI_CHARACTER, destination, tag,    &
     &                      MPI_COMM_WORLD, mpi_error_code)

          ELSE

!............ Loop over all the ranks.

              DO source = first_rank, number_of_processes - 1

!................ Skip receiving from the master rank, because it
!................ shouldn't try to send a message to itself.

                  IF (source /= master_rank) THEN

!.................... Receive the message on this rank (the
!.................... master rank) from the source rank.
                     
!.................... Use length maximum_message_length, which is
!.................... the maximum possible length of a string.
!.................... Note that the message may not have this length;
!.................... for MPI_Recv we need the maximum length that
!.................... the message *could* be.

                      CALL MPI_Recv(                             &
     &                        message, maximum_message_length,   &
     &                        MPI_CHARACTER, source, tag,        &
     &                        MPI_COMM_WORLD, recv_status,       &
     &                        mpi_error_code)

!.................... Output to stdout the message that was received,
!.................... excluding trailing blanks.

                      PRINT *, TRIM(message)

                  END IF !! IF (source /= master_rank)

              END DO !! source

          END IF !! (my_rank /= master_rank)...ELSE

!........ Shut down MPI.

          CALL MPI_Finalize(mpi_error_code)

    END PROGRAM greetings

