!
! nbody.f90
!

      PROGRAM nbody
          USE nbody_particle
          IMPLICIT NONE

          INTEGER,PARAMETER :: memory_success = 0

          CHARACTER (LEN=default_string_length) :: arguments_filename
          REAL,DIMENSION(:,:),ALLOCATABLE :: particle_array
          REAL,DIMENSION(:,:),ALLOCATABLE :: total_force_array
          INTEGER :: timestep
          INTEGER :: memory_status

!........
!........ Determine the name of the file of arguments to input.
!........
          arguments_filename = default_arguments_filename

!........
!........ Input the file of arguments.
!........
          CALL input_arguments(TRIM(arguments_filename))

!........
!........ Allocate and clear an array of particles.
!........
          ALLOCATE(particle_array(number_of_particle_components,  &
     &                            number_of_particles),           &
     &             STAT=memory_status)

          IF (memory_status /= memory_success) THEN

              WRITE (0,*)                                      &
     &            "ERROR: cannot allocate particle_array ",    &
     &            " of size ", number_of_particle_components,  &
     &            " x ", number_of_particles, "."

              CALL die_cleanly()

          END IF !! (memory_status /= memory_success)

          particle_array = 0

!........
!........ Initialize the array of particles randomly.
!........
          CALL ParticleArray_initialize(                         &
     &             particle_array, number_of_particles,          &
     &             domain_size_x, domain_size_y, domain_size_z,  &
     &             mass_maximum)


!........
!........ Allocate and clear an array of total forces.
!........
          ALLOCATE(total_force_array(number_of_force_components,  &
     &                               number_of_particles),        &
     &             STAT=memory_status)

          IF (memory_status /= memory_success) THEN

              WRITE (0,*)                                       &
     &            "ERROR: cannot allocate total_force_array ",  &
     &            " of size ", number_of_force_components,      &
     &            " x ", number_of_particles, "."

              CALL die_cleanly()

          END IF !! (memory_status /= memory_success)

          total_force_array = 0.0

!........
!........ Output the initial array of particles.
!........
          CALL ParticleArray_output(           &
     &             base_filename,              &
     &             particle_array,             &
     &             number_of_particles,        &
     &             0,                          &
     &             number_of_timesteps,        &
     &             timesteps_between_outputs,  &
     &             domain_size_x,              &
     &             domain_size_y,              &
     &             domain_size_z,              &
     &             mass_maximum,               &
     &             time_interval,              &
     &             random_seed_value)

!........
!........ Loop over the timesteps.
!........
          DO timestep = 1, number_of_timesteps

!............
!............ Calculate the total force applied to each particle.
!............
              CALL ForceArray_calculate(                   &
     &                 total_force_array, particle_array,  &
     &                 number_of_particles)

!............
!............ Move each particle based on the total force applied to it.
!............
              CALL ParticleArray_move(                     &
     &                 particle_array, total_force_array,  &
                       number_of_particles, time_interval)

!............
!............ Every so often, output the list of particles.
!............
              IF (MOD(timestep, timesteps_between_outputs) == 0) THEN

                  CALL ParticleArray_output(           &
     &                     base_filename,              &
     &                     particle_array,             &
     &                     number_of_particles,        &
     &                     timestep,                   &
     &                     number_of_timesteps,        &
     &                     timesteps_between_outputs,  &
     &                     domain_size_x,              &
     &                     domain_size_y,              &
     &                     domain_size_z,              &
     &                     mass_maximum,               &
     &                     time_interval,              &
     &                     random_seed_value)

              END IF !! (MOD(timestep, timesteps_between_outputs) == 0)

          END DO !! timestep

!........
!........ If the particles at the last timestep haven't already been
!........ output, output them.
!........
          IF (MOD(number_of_timesteps,  &
     &            timesteps_between_outputs) /= 0) THEN

              CALL ParticleArray_output(      &
     &            base_filename,              &
     &            particle_array,             &
     &            number_of_particles,        &
     &            timestep,                   &
     &            number_of_timesteps,        &
     &            timesteps_between_outputs,  &
     &            domain_size_x,              &
     &            domain_size_y,              &
     &            domain_size_z,              &
     &            mass_maximum,               &
     &            time_interval,              &
     &            random_seed_value)

          END IF !! ((number_of_timesteps % timesteps_between_outputs) /= 0)


!........
!........ Deallocate the array of total forces.
!........
          DEALLOCATE(total_force_array, STAT=memory_status)

          IF (memory_status /= memory_success) THEN

              WRITE (0,*)                                         &
     &            "ERROR: cannot drallocate total_force_array ",  &
     &            " of size ", number_of_force_components,        &
     &            " x ", number_of_particles, "."

              CALL die_cleanly()

          END IF !! (memory_status /= memory_success)

!........
!........ Deallocate the array of particles.
!........
          DEALLOCATE(particle_array, STAT=memory_status)

          IF (memory_status /= memory_success) THEN

              WRITE (0,*)                                        &
     &            "ERROR: cannot deallocate particle_array ",    &
     &            " of size ", number_of_particle_components,    &
     &            " x ", number_of_particles, "."

              CALL die_cleanly()

          END IF !! (memory_status /= memory_success)

      END PROGRAM nbody

!!!!!
!!!!! Get command line arguments.
!!!!!
      SUBROUTINE input_arguments (arguments_filename)
          USE nbody_particle
          IMPLICIT NONE

          CHARACTER(LEN=*),INTENT(IN) :: arguments_filename

          INTEGER,PARAMETER :: input_unit = 10
          INTEGER,PARAMETER :: file_was_closed_code = 0

          number_of_particles       = default_number_of_particles
          domain_size_x             = default_domain_size_x
          domain_size_y             = default_domain_size_y
          domain_size_z             = default_domain_size_z
          time_interval             = default_time_interval
          number_of_timesteps       = default_number_of_timesteps
          timesteps_between_outputs = default_timesteps_between_outputs
          random_seed_value               = default_random_seed_value
          mass_maximum              = default_mass_maximum
          base_filename             = default_base_filename

          IF (LEN_TRIM(arguments_filename) < 1) THEN

              WRITE (0,*) "ERROR in input_arguments: ",  &
     &            "arguments filename doesn't exist"

              CALL die_cleanly()

          END IF !! ((arguments_filename == (char*)NULL) .OR. ... )

          OPEN(FILE=arguments_filename, UNIT=input_unit)

          READ (UNIT=input_unit, FMT=*) number_of_particles

          IF (number_of_particles < 1) THEN

              WRITE (0,*) "ERROR: cannot have ",  &
     &            number_of_particles, " particles!"

              CALL die_cleanly()

          END IF !! (number_of_particles < 1)

          IF (number_of_particles == 1) THEN

              WRITE (0,*)  &
     &            "There is only one particle, therefore no forces."

              CALL die_cleanly()

          END IF !! (number_of_particles == 1)

          READ (UNIT=input_unit, FMT=*) domain_size_x

          IF (domain_size_x <= 0.0) THEN

              WRITE (0,*)                                            &
     &            "ERROR: cannot have a domain whose X dimension ",  &
     &            "has length ", domain_size_x, "!"

              CALL die_cleanly()

          END IF !! (domain_size_x <= 0.0)

          READ (UNIT=input_unit, FMT=*) domain_size_y

          IF (domain_size_y <= 0.0) THEN

              WRITE (0,*)                                            &
     &            "ERROR: cannot have a domain whose Y dimension ",  &
     &            "has length ", domain_size_y, "!"

              CALL die_cleanly()

          END IF !! (domain_size_y <= 0.0)

          READ (UNIT=input_unit, FMT=*) domain_size_z

          IF (domain_size_z <= 0.0) THEN

              WRITE (0,*)                                            &
     &            "ERROR: cannot have a domain whose Z dimension ",  &
     &            "has length ", domain_size_z, "!"

              CALL die_cleanly()

          END IF !! (domain_size_z <= 0.0)

          READ (UNIT=input_unit, FMT=*) time_interval

          IF (time_interval <= 0.0) THEN

              WRITE (0,*)                                    &
     &            "ERROR: cannot have a time interval of ",  &
     &            time_interval, "."

              CALL die_cleanly()

          END IF !! (time_interval <= 0.0)

          READ (UNIT=input_unit, FMT=*) number_of_timesteps

          IF (number_of_timesteps <= 0) THEN

              WRITE (0,*) "ERROR: cannot have ",  &
     &            number_of_timesteps, "!"

              CALL die_cleanly()

          END IF !! (number_of_timesteps <= 0)

          READ (UNIT=input_unit, FMT=*) timesteps_between_outputs

          IF (timesteps_between_outputs <= 0) THEN

              WRITE (0,*)                                            &
     &            "ERROR: cannot have ", timesteps_between_outputs,  &
     &            " timesteps between outputs!"

              CALL die_cleanly()

          END IF !! (timesteps_between_outputs <= 0)

          READ (UNIT=input_unit, FMT=*) random_seed_value

          READ (UNIT=input_unit, FMT=*) mass_maximum

          IF (mass_maximum <= 0.0) THEN

              WRITE (0,*)                                   &
     &            "ERROR: cannot have a maximum mass of ",  &
     &            mass_maximum, "!"

              CALL die_cleanly()

          END IF !! (mass_maximum <= 0.0)

          READ (UNIT=input_unit, FMT="(A)") base_filename

          CLOSE(input_unit)

      END SUBROUTINE input_arguments

